{ Unit: BcPrivileges
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Class to ease setting NT privileges.

  History:

    2001-11-20: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}
{$R-} { Disable range checking because of special AnySizeArrays }

unit BcPrivileges;

interface

uses Windows, BcNT;

type
  TBcSecurityPrivilege = (
    SeCreateTokenPrivilege,
    SeAssignPrimaryTokenPrivilege,
    SeLockMemoryPrivilege,
    SeIncreaseQuotaPrivilege,
    SeUnsolicitedInputPrivilege,
    SeMachineAccountPrivilege,
    SeTcbPrivilege,
    SeSecurityPrivilege,
    SeTakeOwnershipPrivilege,
    SeLoadDriverPrivilege,
    SeSystemProfilePrivilege,
    SeSystemtimePrivilege,
    SeProfileSingleProcessPrivilege,
    SeIncreaseBasePriorityPrivilege,
    SeCreatePagefilePrivilege,
    SeCreatePermanentPrivilege,
    SeBackupPrivilege,
    SeRestorePrivilege,
    SeShutdownPrivilege,
    SeDebugPrivilege,
    SeAuditPrivilege,
    SeSystemEnvironmentPrivilege,
    SeChangeNotifyPrivilege,
    SeRemoteShutdownPrivilege,
    SeUndockPrivilege,
    SeSyncAgentPrivilege,
    SeEnableDelegationPrivilege,
    SeNull);

  TBcSecurityPrivileges = set of TBcSecurityPrivilege;

  TBcSecurityPrivilegeMap = array[TBcSecurityPrivilege] of PChar;

const
  SeBcFirst: TBcSecurityPrivilege = SeCreateTokenPrivilege;
  SeBcLast: TBcSecurityPrivilege = SeEnableDelegationPrivilege;
 
const
  SE_NULL = '';

  BcSecurityPrivilegeMap: TBcSecurityPrivilegeMap = (
    SE_CREATE_TOKEN_NAME,
    SE_ASSIGNPRIMARYTOKEN_NAME,
    SE_LOCK_MEMORY_NAME,
    SE_INCREASE_QUOTA_NAME,
    SE_UNSOLICITED_INPUT_NAME,
    SE_MACHINE_ACCOUNT_NAME,
    SE_TCB_NAME,
    SE_SECURITY_NAME,
    SE_TAKE_OWNERSHIP_NAME,
    SE_LOAD_DRIVER_NAME,
    SE_SYSTEM_PROFILE_NAME,
    SE_SYSTEMTIME_NAME,
    SE_PROF_SINGLE_PROCESS_NAME,
    SE_INC_BASE_PRIORITY_NAME,
    SE_CREATE_PAGEFILE_NAME,
    SE_CREATE_PERMANENT_NAME,
    SE_BACKUP_NAME,
    SE_RESTORE_NAME,
    SE_SHUTDOWN_NAME,
    SE_DEBUG_NAME,
    SE_AUDIT_NAME,
    SE_SYSTEM_ENVIRONMENT_NAME,
    SE_CHANGE_NOTIFY_NAME,
    SE_REMOTE_SHUTDOWN_NAME,
    SE_UNDOCK_NAME,
    SE_SYNC_AGENT_NAME,
    SE_ENABLE_DELEGATION_NAME,
    SE_NULL);

type
  TBcNTSecurityPrivileges = class(TObject)
  private
    FAccessToken: THandle;
  private
    function GetHasPrivileges: TBcSecurityPrivileges;
    function GetPrivilege(APrivilege: TBcSecurityPrivilege): Boolean;
    function GetPrivilegeDescription(
      APrivilege: TBcSecurityPrivilege): string;
    function GetPrivilegeName(APrivilege: TBcSecurityPrivilege): string;
    procedure SetPrivilege(APrivilege: TBcSecurityPrivilege;
      const Value: Boolean);
    function GetPrivilegeValue(APrivilege: TBcSecurityPrivilege): Int64;
  protected
    property AccessToken: THandle read FAccessToken;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function OpenAccessToken: Boolean;
    procedure CloseAccessToken;
    function DisableAll: Boolean;
    function PrivilegeValueToOrd(AValue: Int64): TBcSecurityPrivilege;

    property Privilege[APrivilege: TBcSecurityPrivilege]: Boolean read GetPrivilege write SetPrivilege;
    property PrivilegeName[APrivilege: TBcSecurityPrivilege]: string read GetPrivilegeName;
    property PrivilegeValue[APrivilege: TBcSecurityPrivilege]: Int64 read GetPrivilegeValue;
    property PrivilegeDescription[APrivilege: TBcSecurityPrivilege]: string read GetPrivilegeDescription;
    property HasPrivileges: TBcSecurityPrivileges read GetHasPrivileges;
  published
  end;

function NTSecurityPrivileges: TBcNTSecurityPrivileges;
procedure NTSecurityPrivilegesFree;

implementation

uses SysUtils;

var FBcNTSecurityPrivileges: TBcNTSecurityPrivileges;

function NTSecurityPrivileges: TBcNTSecurityPrivileges;
begin
  if not Assigned(FBcNTSecurityPrivileges) then
    FBcNTSecurityPrivileges := TBcNTSecurityPrivileges.Create;
  Result := FBcNTSecurityPrivileges;
end;

procedure NTSecurityPrivilegesFree;
begin
  FreeAndNil(FBcNTSecurityPrivileges);
end;

{ TBcNTSecurityPrivileges }

procedure TBcNTSecurityPrivileges.CloseAccessToken;
begin
  if FAccessToken <> 0 then
    CloseHandle(FAccessToken);
  FAccessToken := 0;
end;

constructor TBcNTSecurityPrivileges.Create;
begin
end;

destructor TBcNTSecurityPrivileges.Destroy;
begin
  CloseAccessToken;
  inherited Destroy;
end;

function TBcNTSecurityPrivileges.DisableAll: Boolean;
var
  T: Cardinal;
  TP: TTokenPrivileges;
begin
  Result := True;
  if IsNT and OpenAccessToken then
  begin
    Result := AdjustTokenPrivileges(Cardinal(FAccessToken), True, TP, 0, nil, T) = Bool(ERROR_SUCCESS);
    CloseAccessToken;
  end;
end;

function TBcNTSecurityPrivileges.GetHasPrivileges: TBcSecurityPrivileges;
var
  ptPrivileges: PTokenPrivileges;
  dwInfoBufferSize: Cardinal;
  B: Boolean;
  I: Integer;
  SE: TBcSecurityPrivilege;
begin
  if IsNT then
  begin
    Result := [];
    if OpenAccessToken then
    begin
      GetMem(ptPrivileges, 1024);
      try
        B := GetTokenInformation(FAccessToken, TokenPrivileges, ptPrivileges, 1024, dwInfoBufferSize);
        CloseAccessToken;
        if B then
          for I := 0 to ptPrivileges.PrivilegeCount - 1 do
            Include(Result, PrivilegeValueToOrd(ptPrivileges.Privileges[I].Luid));
      finally
        FreeMem(ptPrivileges);
      end;
    end;
  end else // just return all privileges..
    for SE := SeBcFirst to SeBcLast do
      Include(Result, SE);
end;

function TBcNTSecurityPrivileges.GetPrivilege(
  APrivilege: TBcSecurityPrivilege): Boolean;
var
  ptPrivileges: PTokenPrivileges;
  dwInfoBufferSize: Cardinal;
  B: Boolean;
  I: Integer;
begin
  if IsNT then
  begin
    Result := False;
    if OpenAccessToken then
    begin
      GetMem(ptPrivileges, 1024);
      try
        B := GetTokenInformation(FAccessToken, TokenPrivileges, ptPrivileges, 1024, dwInfoBufferSize);
        CloseAccessToken;
        if B then
          for I := 0 to ptPrivileges.PrivilegeCount - 1 do
          begin
            if PrivilegeValueToOrd(ptPrivileges.Privileges[I].Luid) = APrivilege then
            begin
              Result := ptPrivileges.Privileges[I].Attributes = SE_PRIVILEGE_ENABLED;
              Break;
            end;
          end;
      finally
        FreeMem(ptPrivileges);
      end;
    end;
  end else
    Result := True;
end;

function TBcNTSecurityPrivileges.GetPrivilegeDescription(
  APrivilege: TBcSecurityPrivilege): string;
var L, LangID: Cardinal;
begin
  Result := '';
  if IsNT then
  begin
    L := 1024;
    SetLength(Result, L);
    LookupPrivilegeDisplayName(nil, BcSecurityPrivilegeMap[APrivilege],
      PChar(Result), L, LangID);
    Result := PChar(Result); // truncate PChar string
  end;
end;

function TBcNTSecurityPrivileges.GetPrivilegeName(
  APrivilege: TBcSecurityPrivilege): string;
begin
  Result := BcSecurityPrivilegeMap[APrivilege];
end;

function TBcNTSecurityPrivileges.GetPrivilegeValue(
  APrivilege: TBcSecurityPrivilege): Int64;
begin
  Result := 0;
  if IsNT then
    LookupPrivilegeValue(nil, BcSecurityPrivilegeMap[APrivilege], Result)
  else
    Result := Integer(APrivilege);
end;

function TBcNTSecurityPrivileges.OpenAccessToken: Boolean;
begin
  if IsNT then
  begin
    CloseAccessToken;
    Result := OpenThreadToken(GetCurrentThread, TOKEN_ADJUST_PRIVILEGES	or TOKEN_QUERY, True, FAccessToken);
    if not Result then
      if GetLastError = ERROR_NO_TOKEN then
        Result := OpenProcessToken(GetCurrentProcess, TOKEN_ADJUST_PRIVILEGES	or TOKEN_QUERY, FAccessToken);
  end else
    Result := True;
end;

function TBcNTSecurityPrivileges.PrivilegeValueToOrd(
  AValue: Int64): TBcSecurityPrivilege;
var
  S: string;
  L: Cardinal;
  SE: TBcSecurityPrivilege;
begin
  Result := SeNull;
  if IsNT then
  begin
    L := 1024;
    SetLength(S, L);
    if LookupPrivilegeName(nil, AValue, PChar(S), L) then
    begin
      S := PChar(S);
      for SE := SeBcFirst to SeBcLast do
        if BcSecurityPrivilegeMap[SE] = S then
        begin
          Result := SE;
          Break;
        end;
    end;
  end;
end;

procedure TBcNTSecurityPrivileges.SetPrivilege(
  APrivilege: TBcSecurityPrivilege; const Value: Boolean);
const PrivilegeAttributes: array[Boolean] of Cardinal = (0, SE_PRIVILEGE_ENABLED);
var
  T: Cardinal;
  Privilege: TTokenPrivileges;
begin
  { note, AdjustTokenPrivileges fails if requested privilege does not exists. }
  { Read the property value to get the new value and check if it was changed. }
  if IsNT and OpenAccessToken then
  begin
    Privilege.PrivilegeCount := 1;
    Privilege.Privileges[0].Luid := PrivilegeValue[APrivilege];
    Privilege.Privileges[0].Attributes := PrivilegeAttributes[Value];
    AdjustTokenPrivileges(FAccessToken, False, Privilege, 0, nil, T);
    CloseAccessToken;
  end;
end;

initialization

finalization
  NTSecurityPrivilegesFree;

end.
